@@ -1,4 +1,4 @@
-# $Id: Iterator.pm 2684 2007-10-04 12:37:08Z andy $
+# $Id: Iterator.pm 2698 2007-10-04 20:02:43Z andy $
 package Parallel::Iterator;
 
 use warnings;
@@ -11,14 +11,18 @@ use Config;
 
 require 5.008;
 
-our $VERSION = '0.6.0';
+our $VERSION = '0.7.0';
 use base qw( Exporter );
 our @EXPORT_OK = qw( iterate iterate_as_array iterate_as_hash );
 
+use constant IS_WIN32 => ( $^O =~ /^(MS)?Win32$/ );
+
 my %DEFAULTS = (
-    workers => ( $Config{d_fork} ? 10 : 0 ),
-    onerror => 'die',
-    nowarn  => 0,
+    workers => ( ( $Config{d_fork} && !IS_WIN32 ) ? 10 : 0 ),
+    onerror  => 'die',
+    nowarn   => 0,
+    batch    => 1,
+    adaptive => 0,
 );
 
 =head1 NAME
@@ -27,7 +31,7 @@ Parallel::Iterator - Simple parallel execution
 
 =head1 VERSION
 
-This document describes Parallel::Iterator version 0.6.0
+This document describes Parallel::Iterator version 0.7.0
 
 =head1 SYNOPSIS
 
@@ -99,8 +103,8 @@ Now write a function to synthesize a special kind of iterator:
         };
     }
 
-The returned iterator will return each element of the list and then
-undef. Actually it returns both the index I<and> the value of each
+The returned iterator will return each element of the array in turn and
+then undef. Actually it returns both the index I<and> the value of each
 element in the array. Because multiple instances of the transformation
 function execute in parallel the results won't necessarily come back in
 order. The array index will later allow us to put completed items in the
@@ -144,8 +148,74 @@ array instead:
 
     my %done = iterate_as_hash( \&worker, %jobs );
 
+=head2 How It Works
+
+The current process is forked once for each worker. Each forked child is
+connected to the parent by a pair of pipes. The child's STDIN, STDOUT
+and STDERR are unaffected.
+
+Input values are serialised (using Storable) and passed to the workers.
+Completed work items are serialised and returned.
+
 =head2 Caveats
 
+Parallel::Iterator is designed to be simple to use - but the underlying
+forking of the main process can cause mystifying problems unless you
+have an understanding of what is going on behind the scenes.
+
+=head3 Worker execution enviroment
+
+All code apart from the worker subroutine executes in the parent process
+as normal. The worker executes in a forked instance of the parent
+process. That means that things like this won't work as expected:
+
+    my %tally = ();
+    my @r = iterate_as_array( sub {
+        my ($id, $name) = @_;
+        $tally{$name}++;       # might not do what you think it does
+        return reverse $name;
+    }, @names );
+
+    # Now print out the tally...
+    while ( my ( $name, $count ) = each %tally ) {
+        printf("%5d : %s\n", $count, $name);
+    }
+
+Because the worker is a closure it can see the C<%tally> hash from its
+enclosing scope; but because it's running in a forked clone of the parent
+process it modifies its own copy of C<%tally> rather than the copy for
+the parent process.
+
+That means that after the job terminates the C<%tally> in the parent
+process will be empty.
+
+In general you should avoid side effects in your worker subroutines.
+
+=head3 Serialization
+
+Values are serialised using L<Storable> to pass to the worker subroutine
+and results from the worker are again serialised before being passed
+back. Be careful what your values refer to: everything has to be
+serialised. If there's an indirect way to reach a large object graph
+Storable will find it and performance will suffer.
+
+To find out how large your serialised values are serialise one of them
+and check its size:
+
+    use Storable qw( freeze );
+    my $serialized = freeze $some_obj;
+    print length($serialized), " bytes\n";
+
+In your tests you may wish to guard against the possibility of a change
+to the structure of your values resulting in a sudden increase in
+serialized size:
+
+    ok length(freeze $some_obj) < 1000, "Object too bulky?";
+
+See the documetation for L<Storable> for other caveats.
+
+=head3 Performance
+
 Process forking is expensive. Only use Parallel::Iterator in cases where:
 
 =over
@@ -182,18 +252,9 @@ parent use something like this to guard against multiple execution:
         }
     }
 
-=head2 How It Works
-
-The current process is forked once for each worker. Each forked child is
-connected to the parent by a pair of pipes. The child's STDIN, STDOUT
-and STDERR are unaffected.
-
-Input values are serialised (using Storable) and passed to the workers.
-Completed work items are serialised and returned.
-
 =head1 INTERFACE 
 
-=head2 C<< iterate( [ $options ], $trans, $iterator ) >>
+=head2 C<< iterate( [ $options ], $worker, $iterator ) >>
 
 Get an iterator that applies the supplied transformation function to
 each value returned by the input iterator.
@@ -201,13 +262,13 @@ each value returned by the input iterator.
 Instead of an iterator you may pass an array or hash reference and
 C<iterate> will convert it internally into a suitable iterator.
 
-If you are doing this you may with to investigate C<iterate_as_hash> and
+If you are doing this you may wish to investigate C<iterate_as_hash> and
 C<iterate_as_array>.
 
 =head3 Options
 
-A reference to a hash of options may be supplied. The following options
-are supported:
+A reference to a hash of options may be supplied as the first argument.
+The following options are supported:
 
 =over
 
@@ -217,6 +278,40 @@ The number of concurrent processes to launch. Set this to 0 to disable
 forking. Defaults to 10 on systems that support fork and 0 (disable
 forking) on those that do not.
 
+=item C<nowarn>
+
+Normally C<iterate> will issue a warning on systems on which fork is not
+available and fall back to single process mode. This option supresses
+that warning.
+
+=item C<batch>
+
+Ordinarily items are passed to the worker one at a time. If you are
+processing a large number of items it may be more efficient to process
+them in batches. Specify the batch size using this option.
+
+=item C<adaptive>
+
+Extending the idea of batching a number of work items to amortize the
+overhead of passing work to and from parallel workers you may also ask
+C<iterate> to heuristically determine the batch size by setting the
+C<adaptive> option to a numeric value.
+
+The batch size will be computed as
+
+    <number of items seen> / <number of workers> / <adaptive>
+
+A larger value for C<adaptive> will reduce the rate at which the batch
+size increases. Good values tend to be in the range 1 to 2.
+
+You can also specify lower and, optionally, upper bounds on the batch
+size by passing an reference to an array containing ( lower bound,
+growth ratio, upper bound ). The upper bound may be omitted.
+
+    my $iter = iterate(
+        { adaptive => [ 5, 2, 100 ] },
+        $worker, \@stuff );
+
 =item C<onerror>
 
 The action to take when an error is thrown in the iterator. Possible
@@ -232,29 +327,19 @@ of C<$@> thrown.
         $worker,
         \@jobs
     );
+
+The default is 'die'.
     
-=item C<nowarn>
-
-Normally C<iterate> will issue a warning on systems on which fork is not
-available and fall back to single process mode. This option supresses
-that warning.
-
 =back
 
 =cut
 
-sub iterate {
-    my %options = ( %DEFAULTS, %{ 'HASH' eq ref $_[0] ? shift : {} } );
-
-    my $worker = shift;
-    croak "Worker must be a coderef"
-      unless 'CODE' eq ref $worker;
-
+sub _massage_iterator {
     my $iter = shift;
     if ( 'ARRAY' eq ref $iter ) {
         my @ar  = @$iter;
         my $pos = 0;
-        $iter = sub {
+        return sub {
             return if $pos >= @ar;
             my @r = ( $pos, $ar[$pos] );
             $pos++;
@@ -264,18 +349,244 @@ sub iterate {
     elsif ( 'HASH' eq ref $iter ) {
         my %h = %$iter;
         my @k = keys %h;
-        $iter = sub {
+        return sub {
             return unless @k;
             my $k = shift @k;
             return ( $k, $h{$k} );
         };
     }
     elsif ( 'CODE' eq ref $iter ) {
-        # do nothing
+        return $iter;
     }
     else {
         croak "Iterator must be a code, array or hash ref";
     }
+}
+
+sub _nonfork {
+    my ( $options, $worker, $iter ) = @_;
+
+    return sub {
+        while ( 1 ) {
+            if ( my @next = $iter->() ) {
+                my ( $id, $work ) = @next;
+                # dclone so that we have the same semantics as the
+                # forked version.
+                $work = dclone $work if defined $work && ref $work;
+                my $result = eval { $worker->( $id, $work ) };
+                if ( my $err = $@ ) {
+                    $options->{onerror}->( $id, $err );
+                }
+                else {
+                    return ( $id, $result );
+                }
+            }
+            else {
+                return;
+            }
+        }
+    };
+}
+
+# Does this sub look a bit long to you? :)
+sub _fork {
+    my ( $options, $worker, $iter ) = @_;
+
+    my @workers      = ();
+    my @result_queue = ();
+    my $select       = IO::Select->new;
+    my $rotate       = 0;
+
+    return sub {
+        LOOP: {
+            # Make new workers
+            while ( @workers < $options->{workers} && ( my @next = $iter->() ) )
+            {
+
+                my ( $my_rdr, $my_wtr, $child_rdr, $child_wtr )
+                  = map IO::Handle->new, 1 .. 4;
+
+                pipe $child_rdr, $my_wtr
+                  or croak "Can't open write pipe ($!)\n";
+
+                pipe $my_rdr, $child_wtr
+                  or croak "Can't open read pipe ($!)\n";
+
+                if ( my $pid = fork ) {
+                    # Parent
+                    close $_ for $child_rdr, $child_wtr;
+                    push @workers, $pid;
+                    $select->add( [ $my_rdr, $my_wtr, 0 ] );
+                    _put_obj( \@next, $my_wtr );
+                }
+                else {
+                    # Child
+                    close $_ for $my_rdr, $my_wtr;
+
+                    # Worker loop
+                    while ( defined( my $job = _get_obj( $child_rdr ) ) ) {
+                        my $result = eval { $worker->( @$job ) };
+                        my $err = $@;
+                        _put_obj(
+                            [
+                                $err
+                                ? ( 'E', $job->[0], $err )
+                                : ( 'R', $job->[0], $result )
+                            ],
+                            $child_wtr
+                        );
+                    }
+
+                    # End of stream
+                    _put_obj( undef, $child_wtr );
+                    close $_ for $child_rdr, $child_wtr;
+                    exit;
+                }
+            }
+
+            return @{ shift @result_queue } if @result_queue;
+            if ( $select->count ) {
+                eval {
+                    my @rdr = $select->can_read;
+                    # Anybody got completed work?
+                    for my $r ( @rdr ) {
+                        my ( $rh, $wh, $eof ) = @$r;
+                        if ( defined( my $results = _get_obj( $rh ) ) ) {
+                            my $type = shift @$results;
+                            if ( $type eq 'R' ) {
+                                push @result_queue, $results;
+                            }
+                            elsif ( $type eq 'E' ) {
+                                $options->{onerror}->( @$results );
+                            }
+                            else {
+                                die "Bad result type: $type";
+                            }
+
+                            # We operate a strict one in, one out policy
+                            # - which avoids deadlocks. Having received
+                            # the previous result send a new work value.
+                            unless ( $eof ) {
+                                if ( my @next = $iter->() ) {
+                                    _put_obj( \@next, $wh );
+                                }
+                                else {
+                                    _put_obj( undef, $wh );
+                                    close $wh;
+                                    @{$r}[ 1, 2 ] = ( undef, 1 );
+                                }
+                            }
+                        }
+                        else {
+                            $select->remove( $r );
+                            close $rh;
+                        }
+                    }
+                };
+
+                if ( my $err = $@ ) {
+                    # Finish all the workers
+                    _put_obj( undef, $_->[1] ) for $select->handles;
+
+                    # And wait for them to exit
+                    waitpid( $_, 0 ) for @workers;
+
+                    # Rethrow
+                    die $err;
+                }
+
+                redo LOOP;
+            }
+            waitpid( $_, 0 ) for @workers;
+            return;
+        }
+    };
+}
+
+sub _batch_input_iter {
+    my ( $code, $options ) = @_;
+
+    if ( my $adapt = $options->{adaptive} ) {
+        my $workers = $options->{workers} || 1;
+        my $count = 0;
+
+        $adapt = [ 1, $adapt, undef ]
+          unless 'ARRAY' eq ref $adapt;
+
+        my ( $min, $ratio, $max ) = @$adapt;
+        $min = 1 unless defined $min && $min > 1;
+
+        return sub {
+            my @chunk = ();
+
+            # Adapt batch size
+            my $batch = $count / $workers / $ratio;
+            $batch = $min if $batch < $min;
+            $batch = $max if defined $max && $batch > $max;
+
+            while ( @chunk < $batch && ( my @next = $code->() ) ) {
+                push @chunk, \@next;
+                $count++;
+            }
+
+            return @chunk ? ( 0, \@chunk ) : ();
+        };
+    }
+    else {
+        my $batch = $options->{batch};
+
+        return sub {
+            my @chunk = ();
+            while ( @chunk < $batch && ( my @next = $code->() ) ) {
+                push @chunk, \@next;
+            }
+            return @chunk ? ( 0, \@chunk ) : ();
+        };
+    }
+}
+
+sub _batch_output_iter {
+    my $code  = shift;
+    my @queue = ();
+    return sub {
+        unless ( @queue ) {
+            if ( my ( undef, $chunk ) = $code->() ) {
+                @queue = @$chunk;
+            }
+            else {
+                return;
+            }
+        }
+        return @{ shift @queue };
+    };
+    return $code;
+}
+
+sub _batch_worker {
+    my $code = shift;
+    return sub {
+        my ( undef, $chunk ) = @_;
+        for my $item ( @$chunk ) {
+            $item->[1] = $code->( @$item );
+        }
+        return $chunk;
+    };
+}
+
+sub iterate {
+    my %options = ( %DEFAULTS, %{ 'HASH' eq ref $_[0] ? shift : {} } );
+
+    croak "iterate takes 2 or 3 args" unless @_ == 2;
+
+    my @bad_opt = grep { !exists $DEFAULTS{$_} } keys %options;
+    croak "Unknown option(s): ", join( ', ', sort @bad_opt ), "\n"
+      if @bad_opt;
+
+    my $worker = shift;
+    croak "Worker must be a coderef"
+      unless 'CODE' eq ref $worker;
+
+    my $iter = _massage_iterator( shift );
 
     if ( $options{onerror} =~ /^(die|warn)$/ ) {
         $options{onerror} = eval "sub { shift; $1 shift }";
@@ -285,147 +596,25 @@ sub iterate {
       unless 'CODE' eq ref $options{onerror};
 
     if ( $options{workers} > 0 && $DEFAULTS{workers} == 0 ) {
-        warn "Fork not available, falling back to single process mode\n"
+        warn "Fork not available; falling back to single process mode\n"
           unless $options{nowarn};
         $options{workers} = 0;
     }
 
-    if ( $options{workers} == 0 ) {
-        # Non-forking version
-        return sub {
-            while ( 1 ) {
-                if ( my @next = $iter->() ) {
-                    my ( $id, $work ) = @next;
-                    # dclone so that we have the same semantics as the
-                    # forked version.
-                    $work = dclone $work if defined $work && ref $work;
-                    my $result = eval { $worker->( $id, $work ) };
-                    if ( my $err = $@ ) {
-                        $options{onerror}->( $next[0], $err );
-                    }
-                    else {
-                        return ( $next[0], $result );
-                    }
-                }
-                else {
-                    return;
-                }
-            }
-        };
+    my $factory = $options{workers} == 0 ? \&_nonfork : \&_fork;
+
+    if ( $options{batch} > 1 || $options{adaptive} ) {
+        return _batch_output_iter(
+            $factory->(
+                \%options,
+                _batch_worker( $worker ),
+                _batch_input_iter( $iter, \%options )
+            )
+        );
     }
     else {
-
-        # TODO: If we kept track of how many outstanding tasks each worker
-        # had we could load balance more effectively.
-
-        my @workers      = ();
-        my @result_queue = ();
-        my $select       = IO::Select->new;
-
-        # Possibly modify the iterator here...
-
-        return sub {
-            LOOP: {
-                # Make new workers
-                if ( @workers < $options{workers} && ( my @next = $iter->() ) )
-                {
-
-                    my ( $my_rdr, $my_wtr, $child_rdr, $child_wtr )
-                      = map IO::Handle->new, 1 .. 4;
-
-                    pipe $child_rdr, $my_wtr
-                      or croak "Can't open write pipe ($!)\n";
-
-                    pipe $my_rdr, $child_wtr
-                      or croak "Can't open read pipe ($!)\n";
-
-                    $select->add( [ $my_rdr, $my_wtr, 0 ] );
-
-                    if ( my $pid = fork ) {
-                        # Parent
-                        close $_ for $child_rdr, $child_wtr;
-                        push @workers, $pid;
-                        _put_obj( \@next, $my_wtr );
-                    }
-                    else {
-                        # Child
-                        close $_ for $my_rdr, $my_wtr;
-
-                        # Worker loop
-                        while ( defined( my $job = _get_obj( $child_rdr ) ) ) {
-                            my $result = eval { $worker->( @$job ) };
-                            my $err = $@;
-                            _put_obj(
-                                [
-                                    $err
-                                    ? ( 'E', $job->[0], $err )
-                                    : ( 'R', $job->[0], $result )
-                                ],
-                                $child_wtr
-                            );
-                        }
-
-                        # End of stream
-                        _put_obj( undef, $child_wtr );
-                        close $_ for $child_rdr, $child_wtr;
-                        exit;
-                    }
-                }
-
-                return @{ shift @result_queue } if @result_queue;
-                if ( $select->count ) {
-                    eval {
-                        my @rdr = $select->can_read;
-                        # Anybody got completed work?
-                        for my $r ( @rdr ) {
-                            my ( $rh, $wh, $eof ) = @$r;
-                            if ( defined( my $results = _get_obj( $rh ) ) ) {
-                                my $type = shift @$results;
-                                if ( $type eq 'R' ) {
-                                    push @result_queue, $results;
-                                }
-                                elsif ( $type eq 'E' ) {
-                                    $options{onerror}->( @$results );
-                                }
-                                else {
-                                    die "Bad result type: $type";
-                                }
-
-                                unless ( $eof ) {
-                                    if ( my @next = $iter->() ) {
-                                        _put_obj( \@next, $wh );
-                                    }
-                                    else {
-                                        _put_obj( undef, $wh );
-                                        close $wh;
-                                        @{$r}[ 1, 2 ] = ( undef, 1 );
-                                    }
-                                }
-                            }
-                            else {
-                                $select->remove( $r );
-                                close $rh;
-                            }
-                        }
-                    };
-
-                    if ( my $err = $@ ) {
-                        # Finish all the workers
-                        _put_obj( undef, $_->[1] ) for $select->handles;
-
-                        # And wait for them to exit
-                        waitpid( $_, 0 ) for @workers;
-
-                        # Rethrow
-                        die $err;
-                    }
-
-                    redo LOOP;
-                }
-                waitpid( $_, 0 ) for @workers;
-                return;
-            }
-        };
+        # OK. Ready. Let's do it.
+        return $factory->( \%options, $worker, $iter );
     }
 }
 
@@ -461,7 +650,7 @@ returns a reference to the same hash.
 
 For this to work properly the input iterator must return (key, value)
 pairs. This allows the results to be placed in the correct slots in the
-output hash. The simplest way to do this is to pass an hash reference as
+output hash. The simplest way to do this is to pass a hash reference as
 the input iterator:
 
     my %output = iterate_as_hash( \&some_handler, \%input );
@@ -509,7 +698,7 @@ None reported.
 No bugs have been reported.
 
 Please report any bugs or feature requests to
-C<bug-parallel-workers@rt.cpan.org>, or through the web interface at
+C<bug-parallel-iterator@rt.cpan.org>, or through the web interface at
 L<http://rt.cpan.org>.
 
 =head1 AUTHOR
